@@ -1,6 +1,7 @@
 import inspect
 import re
 from functools import wraps
+from typing import Union, Any
 from uuid import uuid4
 
 import pandas
@@ -280,7 +281,7 @@ class DataframeType(BaseType):
             data = data.lower()
         return data
 
-    def replace_prefix(self, value: str) -> str:
+    def replace_prefix(self, value: str) -> Union[str, Any]:
         if isinstance(value, str):
             for prefix, replacement in self.column_prefix_map.items():
                 if value.startswith(prefix):
@@ -291,6 +292,12 @@ class DataframeType(BaseType):
         for i in range(len(values)):
             values[i] = self.replace_prefix(values[i])
         return values
+    
+    def get_comparator_data(self, comparator, value_is_literal: bool = False):
+        if value_is_literal:
+            return comparator
+        else:
+            return self.value.get(comparator, comparator)
 
     @type_operator(FIELD_DATAFRAME)
     def exists(self, other_value):
@@ -304,15 +311,18 @@ class DataframeType(BaseType):
     @type_operator(FIELD_DATAFRAME)
     def equal_to(self, other_value):
         target = self.replace_prefix(other_value.get("target"))
-        comparator = self.replace_prefix(other_value.get("comparator"))
-        results = np.where(self.value.get(target) == self.value.get(comparator, comparator), True, False)
+        value_is_literal = other_value.get("value_is_literal", False)
+        comparator = self.replace_prefix(other_value.get("comparator")) if not value_is_literal else other_value.get("comparator")
+        comparison_data = self.get_comparator_data(comparator, value_is_literal)
+        results = np.where(self.value.get(target) == comparison_data, True, False)
         return pd.Series(results)
 
     @type_operator(FIELD_DATAFRAME)
     def equal_to_case_insensitive(self, other_value):
         target = self.replace_prefix(other_value.get("target"))
-        comparator = self.replace_prefix(other_value.get("comparator"))
-        comparison_data = self.value.get(comparator, comparator)
+        value_is_literal = other_value.get("value_is_literal", False)
+        comparator = self.replace_prefix(other_value.get("comparator")) if not value_is_literal else other_value.get("comparator")
+        comparison_data = self.get_comparator_data(comparator, value_is_literal)
         comparison_data = self.convert_string_data_to_lower(comparison_data)
         results = np.where(self.value.get(target).str.lower() == comparison_data, True, False)
         return pd.Series(results)
@@ -336,26 +346,31 @@ class DataframeType(BaseType):
     @type_operator(FIELD_DATAFRAME)
     def greater_than_or_equal_to(self, other_value):
         target = self.replace_prefix(other_value.get("target"))
-        comparator = self.replace_prefix(other_value.get("comparator"))
-        results = np.where(self.value.get(target) >= self.value.get(comparator, comparator), True, False)
+        value_is_literal = other_value.get("value_is_literal", False)
+        comparator = self.replace_prefix(other_value.get("comparator")) if not value_is_literal else other_value.get("comparator")
+        comparison_data = self.get_comparator_data(comparator, value_is_literal)
+        results = np.where(self.value.get(target) >= comparison_data, True, False)
         return pd.Series(results)
     
     @type_operator(FIELD_DATAFRAME)
     def greater_than(self, other_value):
         target = self.replace_prefix(other_value.get("target"))
-        comparator = self.replace_prefix(other_value.get("comparator"))
-        results = np.where(self.value.get(target) > self.value.get(comparator, comparator), True, False)
+        value_is_literal = other_value.get("value_is_literal", False)
+        comparator = self.replace_prefix(other_value.get("comparator")) if not value_is_literal else other_value.get("comparator")
+        comparison_data = self.get_comparator_data(comparator, value_is_literal)
+        results = np.where(self.value.get(target) > comparison_data, True, False)
         return pd.Series(results)
     
     @type_operator(FIELD_DATAFRAME)
     def contains(self, other_value):
         target = self.replace_prefix(other_value.get("target"))
-        comparator = self.replace_prefix(other_value.get("comparator"))
-        comparison_value = self.value.get(comparator, comparator)
+        value_is_literal = other_value.get("value_is_literal", False)
+        comparator = self.replace_prefix(other_value.get("comparator")) if not value_is_literal else other_value.get("comparator")
+        comparison_data = self.get_comparator_data(comparator, value_is_literal)
         if isinstance(comparator, pandas.core.series.Series):
-            results = np.where(comparison_value.isin(self.value[target]), True, False)
+            results = np.where(comparison_data.isin(self.value[target]), True, False)
         else:
-            results = np.where(self.value[target] == comparison_value, True, False)
+            results = np.where(self.value[target] == comparison_data, True, False)
         return pd.Series(results)
     
     @type_operator(FIELD_DATAFRAME)
@@ -365,8 +380,9 @@ class DataframeType(BaseType):
     @type_operator(FIELD_DATAFRAME)
     def contains_case_insensitive(self, other_value):
         target = self.replace_prefix(other_value.get("target"))
-        comparator = self.replace_prefix(other_value.get("comparator"))
-        comparison_data = self.value.get(comparator, comparator)
+        value_is_literal = other_value.get("value_is_literal", False)
+        comparator = self.replace_prefix(other_value.get("comparator")) if not value_is_literal else other_value.get("comparator")
+        comparison_data = self.get_comparator_data(comparator, value_is_literal)
         comparison_data = self.convert_string_data_to_lower(comparison_data)
         if isinstance(comparator, pandas.core.series.Series):
             results = np.where(comparison_data.isin(self.value[target].str.lower()), True, False)
@@ -381,11 +397,13 @@ class DataframeType(BaseType):
     @type_operator(FIELD_DATAFRAME)
     def is_contained_by(self, other_value):
         target = self.replace_prefix(other_value.get("target"))
+        value_is_literal = other_value.get("value_is_literal", False)
         comparator = other_value.get("comparator")
-        if isinstance(comparator, str):
+        if isinstance(comparator, str) and not value_is_literal:
             # column name provided
             comparator = self.replace_prefix(comparator)
-        results = self.value[target].isin(self.value.get(comparator, comparator))
+        comparison_data = self.get_comparator_data(comparator, value_is_literal)
+        results = self.value[target].isin(comparison_data)
         return pd.Series(results)
     
     @type_operator(FIELD_DATAFRAME)
@@ -396,12 +414,13 @@ class DataframeType(BaseType):
     def is_contained_by_case_insensitive(self, other_value):
         target = self.replace_prefix(other_value.get("target"))
         comparator = other_value.get("comparator", [])
+        value_is_literal = other_value.get("value_is_literal", False)
         if isinstance(comparator, list):
             comparator = [val.lower() for val in comparator]
-        elif isinstance(comparator, str):
+        elif isinstance(comparator, str) and not value_is_literal:
             # column name provided
             comparator = self.replace_prefix(comparator)
-        comparison_data = self.value.get(comparator, comparator)
+        comparison_data = self.get_comparator_data(comparator, value_is_literal)
         if isinstance(comparison_data, pd.core.series.Series):
             comparison_data = comparison_data.str.lower()
         results = self.value[target].str.lower().isin(comparison_data)
@@ -731,6 +750,30 @@ class DataframeType(BaseType):
         results = np.where(target_without_last_row.values == comparator_without_first_row.values, True, False)
         return [*results, pandas.NA]  # appending NA here to make the length of results list the same as length of df
 
+    @type_operator(FIELD_DATAFRAME)
+    def present_on_multiple_rows_within(self, other_value: dict):
+        """
+        The operator ensures that the target is present on multiple rows
+        within a group_by column. The dataframe is grouped by a certain column
+        and the check is applied to each group.
+        """
+        target = self.replace_prefix(other_value.get("target"))
+        min_count: int = other_value.get("comparator") or 1
+        group_by_column = self.replace_prefix(other_value.get("within"))
+        grouped = self.value.groupby(group_by_column)
+        results = grouped.apply(lambda x: self.validate_series_length(x[target], min_count))
+        return pd.Series(results.explode().tolist())
+
+    def validate_series_length(self, ser: pd.Series, min_length: int):
+        if len(ser) > min_length:
+            return [True] * len(ser)
+        else:
+            return [False] * min_length
+
+    @type_operator(FIELD_DATAFRAME)
+    def not_present_on_multiple_rows_within(self, other_value: dict):
+        return ~self.present_on_multiple_rows_within(other_value)
+
     def detect_reference(self, row, value_column, target_column, context=None):
         if context:
             target_data = self.relationship_data.get(row[context], {}).get(row[target_column], pd.Series([]).values)
@@ -739,6 +782,39 @@ class DataframeType(BaseType):
         value = row[value_column]
         return (value in target_data) or (value in target_data.astype(int).astype(str)) or (value in target_data.astype(str))
 
+    @type_operator(FIELD_DATAFRAME)
+    def additional_columns_empty(self, other_value: dict):
+        """
+        The dataframe column might have some additional columns.
+        If the next additional column exists, the previous one cannot be empty.
+        Example:
+            column - TSVAL
+            additional columns - TSVAL1, TSVAL2, ...
+            If TSVAL2 exists -> TSVAL1 cannot be empty.
+            Original column (TSVAL) can be empty.
+
+        The operator extracts these additional columns from the DF
+        and ensures they are not empty.
+        """
+        target: str = self.replace_prefix(other_value.get("target"))
+        regex: str = rf"^{target}\d+$"  # starting from target, ending with integers and nothing is between them
+        df: pd.DataFrame = self.value.filter(regex=regex)
+        # applying a function to each row
+        result: pd.Series = df.apply(lambda row: self.next_column_exists_and_previous_is_null(row), axis=1)
+        return result
+
+    @type_operator(FIELD_DATAFRAME)
+    def additional_columns_not_empty(self, other_value: dict):
+        return ~self.additional_columns_empty(other_value)
+
+    def next_column_exists_and_previous_is_null(self, row: pd.Series) -> bool:
+        row.reset_index(drop=True, inplace=True)
+        for index in row[row.isnull()].index:  # leaving null values only
+            next_position: int = index + 1
+            if next_position < len(row) and row[next_position] is not None:
+                return True
+        return False
+
 @export_type
 class GenericType(SelectMultipleType, SelectType, StringType, NumericType, BooleanType):
 
